"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
class RobustWebSocket {
    constructor(urlOrFactory, onmessage, options) {
        this.urlOrFactory = urlOrFactory;
        this.onmessage = onmessage;
        this.isOpen = false;
        this.closedManually = false;
        this.options = {
            reconnectTimeoutMillis: 5000,
            onOpen: (_) => { },
            onDisconnect: (_) => { }, // eslint-disable-line
        };
        this.pendingOnOpens = [];
        if (options) {
            this.options = Object.assign(Object.assign({}, this.options), options);
        }
        this.init();
    }
    init() {
        this.closedManually = false;
        this.isOpen = false;
        if (typeof this.urlOrFactory === "string") {
            // @ts-ignore  Ignore the error that comes from the compiler not being able to guarantee that
            // this.ws will always be set to the same subtype of WebSocketLike.  We know it will be since
            // this.urlOrFactory can only be set once.
            this.ws = new WebSocket(this.urlOrFactory);
        }
        else {
            this.ws = this.urlOrFactory();
        }
        let reinited = false;
        this.ws.onerror = this.ws.onclose = (ev) => __awaiter(this, void 0, void 0, function* () {
            var _a;
            // Check if we closed manually:
            if (this.closedManually) {
                return;
            }
            console.log(`Websocket to ${(_a = this.ws.url) !== null && _a !== void 0 ? _a : "unknown"} closed or errored (code=${ev.code} reason=${ev.reason}), reconnecting...`);
            const maybePromise = this.options.onDisconnect(this.ws);
            if (maybePromise) {
                yield maybePromise;
            }
            setTimeout(() => {
                if (!reinited) {
                    // Set this in case we get both onerror and onclose for the same socket instance.
                    reinited = true;
                    this.init();
                }
            }, this.options.reconnectTimeoutMillis);
        });
        this.ws.onmessage = this.onmessage;
        if (this.ws.readyState === this.ws.OPEN) {
            this._doOnOpen();
        }
        else {
            this.ws.onopen = () => this._doOnOpen();
        }
    }
    _doOnOpen() {
        return __awaiter(this, void 0, void 0, function* () {
            const maybePromise = this.options.onOpen(this.ws);
            if (maybePromise) {
                yield maybePromise;
            }
            this.isOpen = true;
            for (const f of this.pendingOnOpens) {
                try {
                    f();
                }
                catch (e) {
                    console.error("Error running pending onOpen callback", e);
                }
            }
            this.pendingOnOpens = [];
        });
    }
    _doWhenOpen(f) {
        if (!this.isOpen || this.ws.readyState !== WebSocket.OPEN) {
            this.pendingOnOpens.push(f);
        }
        else {
            f();
        }
    }
    /**
     * Send a message to the server. If the socket is not open, the message will be queued until the
     * connection is reestablished, at which point it will be sent in order.  It is still possible
     * that messages can be lost if the connection is broken in transit or right before.
     */
    send(msg) {
        this._doWhenOpen(() => {
            this.ws.send(msg);
        });
    }
    // Close the connection to the server.  It will not be reestablished in this case.
    close(code, reason) {
        this.closedManually = true;
        if (this.ws.close) {
            this.ws.close(code, reason);
        }
    }
}
exports.default = RobustWebSocket;
//# sourceMappingURL=index.js.map